@@ -1,5 +1,5 @@
-from itertools import product, islice
 import sys
+from warnings import warn
 import six
 from .arbitraries.arbitrary import get_arbitrary
 from .arbitraries.sequences import dict_fixkeys
@@ -34,10 +34,11 @@ class CheckError(Exception):
     cause = None
 
 
-def get_CheckError(data, cause):
+def get_CheckError(data, noshrink_data, cause):
     class CheckError_(CheckError, type(cause)):
-        def __init__(self, data, cause):
+        def __init__(self, data, noshrink_data, cause):
             self.test_data = data
+            self.noshrink_data = noshrink_data
             self.cause = cause
 
         def __str__(self):
@@ -49,12 +50,16 @@ def get_CheckError(data, cause):
                     msg += ": " + cause
 
             res = [msg]
+            res.append('')
             res.append("Failure encountered for data:")
             res += ["  %s = %r" % i for i in sorted(self.test_data.items())]
+            res.append('')
+            res.append("Before shrinking:")
+            res += ["  %s = %r" % i for i in sorted(self.noshrink_data.items())]
             return '\n'.join(res)
 
     CheckError_.__name__ = 'CheckError'
-    return CheckError_(data, cause)
+    return CheckError_(data, noshrink_data, cause)
 
 
 class ReturnedFalse(Exception):
@@ -76,7 +81,7 @@ class Result(object):
 
     def __init__(self,
                  verdict, passed, datas,
-                 data=None, exception=None, traceback=None):
+                 data=None, noshrink_data=None, exception=None, traceback=None):
         #: ``'OK'`` if passed and ``'FAIL'`` if not
         self.verdict = verdict
         #: ``True`` if passed and ``False`` if not (opposite to :attr:`failed`)
@@ -88,6 +93,8 @@ class Result(object):
         #: Data on which the tested function failed (as a :mod:`dict`),
         #: or ``None`` if passed
         self.data = data
+        #: Same as :attr:`data`, but before shrinking.
+        self.noshrink_data = noshrink_data
         #: Exception raised by the function, or :exc:`ReturnedFalse` if function
         #: returned ``False``, or ``None`` if passed
         self.exception = exception
@@ -99,11 +106,14 @@ class Result(object):
         return Result('OK', True, datas)
 
     @classmethod
-    def fail(cls, datas, data, exception, traceback):
-        return Result('FAIL', False, datas, data, exception, traceback)
+    def fail(cls, datas, data, noshrink_data, exception, traceback):
+        return Result('FAIL', False,
+                      datas, data, noshrink_data,
+                      exception, traceback)
 
     def reraise(self):
-        exception = get_CheckError(self.data, self.exception)
+        exception = get_CheckError(self.data, self.noshrink_data,
+                                   self.exception)
         six.reraise(type(exception), exception, self.traceback)
 
     def __bool__(self):
@@ -115,10 +125,12 @@ class Result(object):
 
 
 class Checker(object):
-    def __init__(self, func, args, ntests, ignore_return=False):
+    def __init__(self, func, args, ntests, nshrinks, nassume, ignore_return=False):
         self.func = func
         self.args = args
-        self.ntests = ntests
+        self.ntests = ntests or 500
+        self.nshrinks = nshrinks or 1000
+        self.nassume = nassume or 1000
         self.ignore_return = ignore_return
 
         arg_arbs = {argname: get_arbitrary(self.args[argname])
@@ -133,20 +145,29 @@ class Checker(object):
 
 
     def full_shrink(self, kwargs):
-        while True:
+        n = 0
+        while n < self.nshrinks:
             for shr_kwargs in self.arb.shrink(kwargs):
+                n += 1
                 try:
                     self.eval_func(shr_kwargs)
+                except AssumptionError:
+                    continue
                 except Exception as e:
                     kwargs = shr_kwargs
                     break
+                if n > self.nshrinks:
+                    break
             else:
                 break
+        else:
+            warn('Shrink did not stop after %d iterations.' % self.nshrinks)
         return kwargs
 
 
     def for_all(self):
         gen_serial = self.arb.gen_serial()
+
         def generate(n):
             if n < self.ntests / 2:
                 try:
@@ -167,30 +188,35 @@ class Checker(object):
             except AssumptionError:
                 del datas[-1]
                 assume_fail_cnt += 1
-                if assume_fail_cnt > 10000:
-                    raise RuntimeError('Could not satisfy assume() after 10000 of test cases.')
+                if assume_fail_cnt > self.nassume:
+                    raise RuntimeError(
+                        'Could not satisfy assume() after %d of test cases.' % self.nassume)
                 continue
             except Exception as e:
-                kwargs = self.full_shrink(kwargs)
+                shrinked_kwargs = self.full_shrink(kwargs)
                 try:
-                    self.eval_func(kwargs)
+                    self.eval_func(shrinked_kwargs)
                 except Exception as e:
                     _, _, traceback = sys.exc_info()
-                    return Result.fail(datas, kwargs, e, traceback)
+                    return Result.fail(datas, shrinked_kwargs, kwargs, e,
+                                       traceback)
             assume_fail_cnt = 0
             n += 1
 
         return Result.ok(datas)
 
 
-def for_all(func, _ntests=500, **kwargs):
-    """ Test the function ``func`` on ``_ntests`` tests.
+def for_all(func, **kwargs):
+    """ for_all(func, ntests=None, nshrinks=None, nassume=None)
+    Test the function ``func`` on ``_ntests`` tests.
 
     For usage examples see :ref:`property-checking`.
 
-    :param _ntests: number of tests
     :param kwargs: function arguments specification
     :returns: information about the tests run
     :rtype: :class:`Result`
     """
-    return Checker(func, kwargs, _ntests).for_all()
+    ntests = kwargs.get('ntests', None)
+    nshrinks = kwargs.get('nshrinks', None)
+    nassume = kwargs.get('nassume', None)
+    return Checker(func, kwargs, ntests, nshrinks, nassume).for_all()
